/*
 * Copyright 2015-2017, Thomas Maier-Komor
 * LICENSE: GPLv3, see file LICENSE for details
 */

#include <arpa/inet.h>
#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <inttypes.h>
#include <signal.h>
#include <stdlib.h>
#include <sys/resource.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/sysinfo.h>
#include <sys/types.h>
#include <sys/utsname.h>
#include <unistd.h>

#ifdef WFC
#include "hostscope_wfc.h"
#else
#include "hostscope.pb.h"
#endif

#include "config.h"
#include "common.h"
#include "linuxstats.h"
#include "log.h"
#include "support.h"


using namespace std;


extern int MaxFd;
static int MinUserUid;
static useconds_t Sleeptime = INTERVALTIME;
static int Sock = -1;
static struct sockaddr_in MA;
static bool Daemon = true, Localmode = false;
static bool ReadDisks = true, ReadNets = true, ReadProcs = true;


static void open_mcast_socket()
{
	Sock = socket(AF_INET, SOCK_DGRAM, 0);
	if (-1 == Sock)
		error("could not create socket: %s\n",strerror(errno));
	bzero(&MA,sizeof(MA));
	MA.sin_family = AF_INET;
	MA.sin_port = htons(Port);
#if SOLARIS >= 8
	if (-1 == inet_pton(AF_INET, McastAddr, &MA.sin_addr))
		error("could not set multicast address: %s\n",strerror(errno));
#else
	MA.sin_addr.s_addr = inet_addr(McastAddr);
	if ((in_addr_t)(-1) == MA.sin_addr.s_addr)
		error("could not set multicast address\n");
#endif
}


static void sendall(const char *buf, size_t s)
{
	dbug("sendall(%p,%u)\n",buf,s);
	if (Localmode) {
		if (-1 == write(STDOUT_FILENO,buf,s)) {
			if (errno == EPIPE)
				exit(EXIT_SUCCESS);
			error("could not transmit status to client: %s\n",strerror(errno));
		}
		return;
	}
	int err;
	do {
		err = sendto(Sock, buf, s, 0, (const struct sockaddr *) &MA, sizeof(MA));
		if (err < 0) {
			error("could not multicast message: %s; trying to reopen socket\n",strerror(errno));
			close(Sock);
			sleep(1);
			open_mcast_socket();
		}
	} while (err < 0);
}


void sigHandler(int sig)
{
	assert((sig == SIGINT) || (sig == SIGTERM));
	unlink(LOCKFILE);
	exit(EXIT_SUCCESS);
}


static void set_hostinfo(HostScope_HostInfo &hinfo, HostScope_LoadAvg &lavg)
{
	static struct utsname utsname;
	static int ret = uname(&utsname);
	assert(ret == 0);
	hinfo.set_hostname(utsname.nodename);
	hinfo.set_osname(utsname.sysname);
	if (utsname.release && utsname.release[0])
		hinfo.set_osrelease(utsname.release);
	if (utsname.version && utsname.version[0])
		hinfo.set_osversion(utsname.version);
	if (utsname.machine && utsname.machine[0])
		hinfo.set_machtype(utsname.machine);
	struct sysinfo si;
	sysinfo(&si);
	hinfo.set_uptime(si.uptime);
	double l[3];
	l[0] = ((double)si.loads[0]/65536.0);
	l[1] = ((double)si.loads[1]/65536.0);
	l[2] = ((double)si.loads[2]/65536.0);
	dbug("loadav1 %u => %g\n",si.loads[0],l[0]);
	dbug("loadav5 %u => %g\n",si.loads[1],l[1]);
	dbug("loadav15 %u => %g\n",si.loads[2],l[2]);
	lavg.set_loadav1m((uint32_t)(l[0]*1000.0));
	lavg.set_loadav5m((uint32_t)(l[1]*1000.0));
	lavg.set_loadav15m((uint32_t)(l[2]*1000.0));
}


static int read_minuseruid()
{
	int fd = open("/etc/login.defs",O_RDONLY);
	if (fd == -1) {
		warning("unable to open /etc/login.defs: %s\n",strerror(errno));
		return -1;
	}
	struct stat st;
	if (-1 == fstat(fd,&st)) {
		close(fd);
		warning("unable to stat /etc/login.defs: %s\n",strerror(errno));
		return -1;
	}
	char *buf = (char *)malloc(st.st_size + 1);
	assert(buf);
	if (-1 == read(fd,buf,st.st_size)) {
		close(fd);
		warning("unable to read /etc/login.defs: %s\n",strerror(errno));
		return -1;
	}
	close(fd);
	char * at = 0;
	do {
		at = strstr(at ? at+1 : buf,"UID_MIN");
	} while ((at != 0) && (at != buf) && (at[-1] != '\n'));
	if (at == 0)
		return -1;
	return strtol(at+8,0,10);
}


static void usage(const char *m, int r)
{
	printf(
	"hostscoped [options]\n"
	"Options:\n"
	"-h        : display this help screen and exit\n"
	"-k        : kill hostscope's daemon\n"
	"-L        : operate in pipe mode\n"
	"-l <log>  : create log file <log>\n"
	"-m <a:p>  : set multicast group address and port (optional) to <a:p>\n"
	"-p <port> : set multicast group port to <port>\n"
	"-f        : run in foreground mode - i.e. do not run as a daemon\n"
	"-d        : disable disk stats gathering\n"
	"-e        : enable effective uid determination\n"
	"-u        : disable user process statistics gathering\n"
	"-i        : disable user process I/O statistics gathering\n"
	"-n        : omit network statistics\n"
	"-M        : include shared memory info from /proc/<pid>/statm\n"
	"-v <num>  : set verbosity to num, where 0=silent and 5=debugging\n"
	"-V        : display version number and exit\n");
	if (m[0]) 
		printf("\n%s\n",m);
	exit(r);
}


static void parse_argv(char a)
{
	int v;
	switch (a) {
	default:
		abort();
		break;
	case 'd':
		ReadDisks = false;
		break;
	case 'e':
		ReadEuid = true;
		break;
	case 'f':
		Daemon = false;
		break;
	case 'h':
		usage("",EXIT_SUCCESS);
		break;
	case 'i':
		ReadProcIO = false;
		break;
	case 'k':
		kill_daemon();
		break;
	case 'L':
		Localmode = true;
		break;
	case 'l':
		setLogfile(optarg);
		break;
	case 'M':
		ReadStatm = true;
		break;
	case 'm':
		setMulticastAddr(optarg);
		break;
	case 'n':
		ReadNets = false;
		break;
	case 'p':
		setMulticastPort(optarg);
		break;
	case 's':
		log2syslog(DAEMONEXE);
		break;
	case 'u':
		ReadProcs = false;
		break;
	case 'v':
		if (1 == sscanf(optarg,"%d",&v)) {
			if (v < 0 || v > 5)
				error("value out of range for option -v\n");
			else
				Verbose = v;
			info("verbose set to %d\n",Verbose);
		} else
			error("invalid argument for option -v\n");
		break;
	case 'V':
		print_version();
		exit(EXIT_SUCCESS);
	case ':':
		usage("missing argument",EXIT_FAILURE);
		break;
	case '?':
		usage("",EXIT_FAILURE);
		break;
	}
}


static int create_lockfile()
{
	int pidf = open(LOCKFILE,O_RDWR|O_TRUNC|O_CREAT|O_EXCL,0666);
	while (pidf == -1) {
		dbug("start_daemon(): lock file exists\n");
		if (errno != EEXIST)
			fatal("unable to create lock file " LOCKFILE ": %s\n",strerror(errno));
		pidf = open(LOCKFILE,O_RDONLY,0666);
		if (pidf == -1)
			fatal("unable to open existing lock file " LOCKFILE ": %s\n",strerror(errno));
		char buf[64];
		int n = read(pidf,buf,sizeof(buf));
		if (n == 0) {
			sleep(1);
			n = pread(pidf,buf,sizeof(buf),0);
			if (n == 0) {
				warning("removing empty stale pid file\n");
				close(pidf);
				int r = unlink(LOCKFILE);
				if (r == -1)
					fatal("unable to remove stale and empty pid file: %s\n",strerror(errno));
				pidf = open(LOCKFILE,O_RDWR|O_TRUNC|O_CREAT|O_EXCL,0666);
				continue;
			}
		}
		close(pidf);
		if (n == -1)
			fatal("unable to read existing lock file " LOCKFILE ": %s\n",strerror(errno));
		long xpid = strtol(buf,0,10);
		dbug("existing daemon pid %d\n",xpid);
		assert(xpid != 0);
		snprintf(buf,sizeof(buf),"/proc/%ld/comm",xpid);
		int exefd = open(buf,O_RDONLY);
		if (-1 != exefd) {	// pid exists
			n = read(exefd,buf,sizeof(buf));
			assert(n > 0);
			close(exefd);
			buf[n] = 0;
			dbug("existing daemon comm(%d): '%s'\n",n,buf);
			if (0 == strcmp(buf,"hostscoped\n")) { // daemon already running
				fatal("daemon already running with pid %d\n",xpid);
			}
			warning("daemon's pid has unexpected comm: %s\n",buf);
		}
		info("removing stale lock file " LOCKFILE "\n");
		if (-1 == unlink(LOCKFILE))
			fatal("unable to remove stale lock file " LOCKFILE ": %s\n",strerror(errno));
		dbug("retrying to create lockfile\n");
		pidf = open(LOCKFILE,O_RDWR|O_TRUNC|O_CREAT|O_EXCL,0666);
	}
	return pidf;
}


static void start_daemon()
{
	int pidf = create_lockfile();
	int pid = fork();
	if (-1 == pid)
		fatal("daemon creation failed while forking: %s\n",strerror(errno));
	if (pid != 0)
		exit(EXIT_SUCCESS);
	int cpid = fork();
	if (-1 == cpid)
		fatal("daemon creation failed while forking: %s\n",strerror(errno));
	if (cpid == 0) {
		close(pidf);
		return;
	}
	char buf[16];
	size_t l = snprintf(buf,sizeof(buf),"%u\n",cpid);
	int n = write(pidf,buf,l);
	assert(n != -1);
	(void) close(pidf);
	exit(EXIT_SUCCESS);
}


static void gather_data(HostScope &t)
{
	TimeStat(__FUNCTION__);
	static unsigned i = 0;
	struct timespec ts;
	clock_gettime(CLOCK_MONOTONIC,&ts);

	HostScope_MetaData *md = t.mutable_metadata();
	md->set_protocol_version(PROTOCOL_VERSION);
	md->set_hostscope_version(Version);
	time_t now = time(0);
	md->set_timestamp(now);
	HostScope_HostInfo &hi = *t.mutable_hostinfo();
	hi.set_minuseruid(MinUserUid);
	HostScope_LoadAvg &la = *t.mutable_loadavg();
	set_hostinfo(hi,la);
	if (ReadDisks)
		read_diskinfo(&t);
	read_meminfo(&t);
#ifndef __CYGWIN__
	read_vmstat(&t);
#endif
	read_cpuinfo(&t);
	if (ReadNets)
		read_networks(&t);
	if (ReadProcs)
		read_procs(&t);
	read_powersupplies(&t);
	++i;
}


static void sendout(HostScope &t)
{
	TimeStat(__FUNCTION__);
#ifdef WFC
	static size_t bufsize = ((t.calcSize()*2)/4096+1)*4096;
	static uint8_t *buf = (uint8_t *) malloc(bufsize);

	size_t n;
	n = t.toMemory(buf,bufsize);
	if (n == 0) {
		free(buf);
		bufsize = ((t.calcSize()*2)/4096+1)*4096;
		buf = (uint8_t *) malloc(bufsize);
		n = t.toMemory(buf,bufsize);
		assert(n != 0);
	}
#ifdef SANITY_CHECK
	HostScope check;
	check.fromMemory(buf,l);
	assert(t == check);
#endif
	sendall((const char *)buf,n);
#else	// original protocol buffers
	string str;
	t.SerializeToString(&str);
	sendall(str.c_str(),str.size());
#endif
}


int main(int argc, char *argv[])
{
#ifdef GOOGLE_PROTOBUF_VERIFY_VERSION
	GOOGLE_PROTOBUF_VERIFY_VERSION;
#endif
	int i;
	while ((i = getopt(argc,argv,"defhikLl:Mm:np:suVv:")) != -1)
		parse_argv(i);
	if (Daemon)
		setLogfile(DAEMONLOG);
	if (-1 == chdir(RUNDIR))
		fatal("Could not change directory to " RUNDIR ": %s\n",strerror(errno));
	if (Daemon) {
		start_daemon();
	} else {
		int pidf = create_lockfile();
		if (-1 == pidf)
			fatal("daemon already running\n");
		char buf[16];
		sprintf(buf,"%d\n",getpid());
		int n = write(pidf,buf,strlen(buf));
		if (n < 0)
			error("could not write pid to lockfile " LOCKFILE ": %s\n",strerror(errno));
		close(pidf);
		signal(SIGTERM,sigHandler);
		signal(SIGINT,sigHandler);
	}
	struct rlimit lim;
	if (0 == getrlimit(RLIMIT_NOFILE,&lim))
		MaxFd = lim.rlim_cur;
	ProcFd = open("/proc",O_RDONLY);
	if (-1 == ProcFd)
		fatal("unable to open /proc: %s\n",strerror(errno));
	signal(SIGSEGV,sigSegVhandler);
	signal(SIGABRT,sigSegVhandler);
	HostScope t;
	open_mcast_socket();
	init_tsensors();
	init_powersupplies(&t);

	MinUserUid = read_minuseruid();
	if (MinUserUid <= 0) {
		warning("unable to determine minmum user uid. using default %d\n",DEFAULT_MINUSERUID);
		MinUserUid = DEFAULT_MINUSERUID;
	}
	info("minimum user uid: %d\n",DEFAULT_MINUSERUID);
	gather_data(t);
	usleep(Sleeptime/10);
	for (;;) {
		gather_data(t);
		sendout(t);
		usleep(Sleeptime);
	}
}
